/* This program is free software; you can redistribute it and/or */
/* modify it under the terms of the GNU General Public License */
/* as published by the Free Software Foundation; either version 2 */
/* of the License, or (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#if WITH_INOTIFY
#include "inotify.h"
#include "inotify-syscalls.h"
#endif

#include "inotify_edit.h"
#include "gutil.h"
#include "command_tab.h"
#include "appconfig.h"


extern GtkWidget *window1;

static GHashTable   *itable;

#if WITH_INOTIFY
static gint         infd=-1;
static GIOChannel   *ioc;
#endif

typedef struct instruct_ {
  gint         fd;
  gchar        *filename;
  gchar        *args[10];
  GtkWidget    *page;
  GtkWidget    *pagelabel;
  GtkWidget    *button;
  GtkWidget    *textview;
  GPid         pid;
} instruct;


static void update_buffer_on_notify(GtkWidget *textview,
				    gchar     *filename)
{
	GtkTextBuffer  *buffer;
	GtkTextIter    start, end;
	gint           file, r;
	gchar          msg[256], buf[8196];

	file = open((gchar *)filename, O_RDONLY, "r"); 
  
	if( file == -1) {
  
		g_snprintf(msg, sizeof(msg), 
			   "Could not open '%s' for inotify update, errno: %d",
			   (gchar *)filename, errno);
		gtk_label_set_text(GTK_LABEL(lw("statusLabel")), msg);
		return;
	}

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textview));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	gtk_text_buffer_delete(buffer, &start, &end);

	while( (r = read(file, buf, sizeof(buf)-1)) > 0 )
		gtk_text_buffer_insert(buffer, &start, buf, r);
    
	close(file);
	g_snprintf(msg, sizeof(msg), "inotify updated '%s'", filename);
	gtk_label_set_text(GTK_LABEL(lw("statusLabel")), msg);
}





#if WITH_INOTIFY
static gboolean f_notify(GIOChannel    *source, 
			 GIOCondition  condition,
			 gpointer      data)
{

	gint      buflen = 1024 * (sizeof(struct inotify_event) + 16);
	gint      i=0, len;
	gchar     buf[buflen];
	instruct  *is;

	len = read(infd, buf, buflen);

	if (len < 0) {
		fprintf(stderr,"inotify_edit: error on read of inotify fd\n");
		return FALSE;
	}
	else {
		while (i < len) {
			struct inotify_event     *event;
      
			event = (struct inotify_event *) &buf[i];

			is = (instruct *) g_hash_table_lookup(itable, &event->wd);

			if(event->mask & IN_MODIFY) {

#if DEBUG
				fprintf(stderr,"inotify_edit: IN_MODIFY: %s\n", is->filename);
#endif
				update_buffer_on_notify(is->textview, is->filename);
			}

#if DEBUG
			if(event->mask & IN_CLOSE_WRITE)
				fprintf(stderr,"inotify_edit: IN_CLOSE_WRITE: %s\n", is->filename);
			if(event->mask & IN_CLOSE_NOWRITE)
				fprintf(stderr,"inotify_edit:IN_CLOSE_NOWRITE: %s\n", is->filename);
#endif

			i += sizeof (struct inotify_event) + event->len;
		}
	}


	return TRUE;
}
#endif

static void editor_exit(GPid gpid, gint status, gpointer data) 
{
	gchar    msg[512];
	instruct *is = (instruct *) data;


#if WITH_INOTIFY
	inotify_rm_watch(infd, is->fd);  
#endif

	update_buffer_on_notify(is->textview, is->filename);
	unlink(is->filename);

	g_snprintf(msg, sizeof(msg), "<b>Program editing %s exited</b>", is->filename);
	gtk_label_set_markup(GTK_LABEL(lw("statusLabel")), msg);

#if DEBUG
	fprintf(stderr, "inotify_edit: Editor child process %d exited\n", is->pid);
#endif

	gtk_widget_show(is->button);
	gtk_widget_set_sensitive(is->textview, TRUE);

	g_free(is->filename);
	g_free(is);
}


void inotify_edit_open(guint editor) 
{
	gchar            *sql, msg[256], basename[64];
	gint             file;
	instruct         *is;
	gchar            argbuf[256];
	const gchar      *ext_editor, *tmp_dir;
	GError           *gerror=0;


	if( (ext_editor=config_get("external_editor")) == NULL) {
		g_snprintf(msg, sizeof(msg), "could not lookup up 'external_editor' in config_hash");
		gtk_label_set_text(GTK_LABEL(lw("statusLabel")), msg);
		return;
	}
  
#if WITH_INOTIFY
	if(infd == -1) {

		infd = inotify_init();

		if(infd == -1) {
			GtkWidget *dialog, *label;

			dialog = gtk_dialog_new_with_buttons ((const gchar *)"Error initializing inotify",
							      GTK_WINDOW(window1),
							      GTK_DIALOG_DESTROY_WITH_PARENT,
							      GTK_STOCK_OK,
							      GTK_RESPONSE_NONE,
							      NULL);
			label = gtk_label_new("Call to inotify_init() failed\nGTKO was configured with '--with-inotify'\nbut the inotify interface could not be initialized.\n\nMake sure:\n  - the kernel version is at least 2.6.13\n  - inotify support is enabled in the kernel\n        Filesystems-->Inotify file change notification support\n");
			g_signal_connect_swapped (dialog,
						  "response", 
						  G_CALLBACK (gtk_widget_destroy),
						  dialog);

			gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),label);
			gtk_widget_show_all (dialog);
			return;
		}
		else {

			ioc = g_io_channel_unix_new(infd);
			g_io_add_watch(ioc, G_IO_IN, f_notify, (gpointer) NULL);
		}
	}
#endif   

	if(!itable)
		itable = g_hash_table_new(g_int_hash, g_int_equal);

	is = g_malloc(sizeof(instruct));
	g_snprintf(basename, sizeof(basename), "%s_%u.sql", gtk_label_get_text( GTK_LABEL(command_tab_get_page_label())), g_random_int());

#if WIN32
	is->filename  = g_build_filename("C:", "TEMP", basename, NULL);
#else
	is->filename  = g_build_filename("/tmp", basename, NULL);
#endif

	is->page      = command_tab_get_page();
	is->pagelabel = command_tab_get_page_label();
	is->button    = command_tab_get_page_button();
	is->textview  = GTK_WIDGET(command_tab_get_current_textview());
 
	file = open(is->filename, O_WRONLY | O_CREAT | O_TRUNC, "w");

	if( file == -1 ) {

		g_snprintf(msg, sizeof(msg), "<span color=\"red3\">Could not open inotify temp file '%s' for write: %d</span>", is->filename, errno);
		gtk_label_set_markup(GTK_LABEL(lw("statusLabel")), msg);

		g_free(is->filename);
		g_free(is);
		return;
	}
  
	sql = getCommandText(-1);
	write(file, sql, strlen(sql));
	close(file);
	g_free(sql);

	chmod(is->filename, S_IRUSR | S_IWUSR );

#if WITH_INOTIFY  
	is->fd = inotify_add_watch(infd, is->filename, IN_MODIFY | IN_CLOSE_WRITE | IN_CLOSE_NOWRITE );
 
	if(is->fd == -1) {
    
		g_snprintf(msg, sizeof(msg), "<span color=\"red3\">inotify_add_watch(%d, %s, IN_MODIFY | IN_CLOSE) failed: %d</span>", is->fd, is->filename, errno);
		gtk_label_set_markup(GTK_LABEL(lw("statusLabel")), msg);  
		return;
	}

	g_snprintf(msg, sizeof(msg), "<b>Added inotify watch for external editor file '%s': %d</b>", is->filename, infd);
	gtk_label_set_markup(GTK_LABEL(lw("statusLabel")), msg);  
#endif


	g_snprintf(argbuf, sizeof(argbuf), "%s", ext_editor);

#if DEBUG
	fprintf(stderr, "inotify_edit: spawning '%s'\n", argbuf);
#endif

	is->args[0] = g_strdup(argbuf);
	is->args[1] = is->filename;
	is->args[2] = NULL;

	g_hash_table_insert(itable, &is->fd, (gpointer) is);

	tmp_dir = g_get_tmp_dir();
	if(g_spawn_async_with_pipes(tmp_dir,
				    is->args,
				    NULL,
				    G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
				    NULL,
				    (gpointer) NULL,
				    &is->pid,
				    NULL,
				    NULL,
				    NULL,
				    &gerror) == TRUE) {
		g_child_watch_add(is->pid, editor_exit, (gpointer) is);
  
		gtk_widget_hide(is->button);
		gtk_widget_set_sensitive(is->textview, FALSE);
	}
	else {
#if WITH_INOTIFY
		inotify_rm_watch(infd, is->fd);
#endif
		unlink(is->filename);

		g_snprintf(msg, sizeof(msg), "<span color=\"red3\">Could not execute '%s': %s</span>", argbuf, gerror->message);
		gtk_label_set_markup(GTK_LABEL(lw("statusLabel")), msg);
	}
}


