/* This program is free software; you can redistribute it and/or */
/* modify it under the terms of the GNU General Public License */
/* as published by the Free Software Foundation; either version 2 */
/* of the License, or (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <gutil.h>
#include <glib.h>

#include "statement.h"


static gchar *command[] = {
  "SELECT",
  "INSERT",
  "UPDATE",
  "DELETE",
  "ALTER",
  "ANALYZE",
  "ASSOCIATE",
  "AUDIT",
  "CALL",
  "COMMENT",
  "COMMIT",
  "CREATE",
  "DISASSOCIATE",
  "DROP",
  "EXPLAIN",
  "FLASHBACK",
  "GRANT",
  "LOCK",
  "MERGE",
  "NOAUDIT",
  "PURGE",
  "RENAME",
  "REVOKE",
  "ROLLBACK",
  "SAVEPOINT",
  "TRUNCATE",
  "WITH"
};

static gchar *plsql[] = {
  "FUNCTION",
  "PROCEDURE",
  "PACKAGE",
  "JAVA",
  "OR"
};

static gchar *ignore[] = {
  "SET",
  "COLUMN",
  "FORMAT",
  "BREAK",
  "COMPUTE",
  "--",
  "REMARK",
  "ACCEPT",
  "APPEND",
  "ATTRIBUTE",
  "BTITLE",
  "CHANGE",
  "CLEAR",
  "CONNECT",
  "COPY",
  "DEFINE",
  "DESCRIBE",
  "DISCONNECT",
  "EDIT",
  "EXECUTE",
  "EXIT",
  "GET",
  "HELP",
  "HOST",
  "INPUT",
  "LIST",
  "PASSWORD",
  "PAUSE",
  "PRINT",
  "PROMPT",
  "RECOVER",
  "REMARK",
  "REPFOOTER",
  "REPHEADER",
  "RUN",
  "SAVE",
  "SET",
  "SHOW",
  "SPOOL",
  "STORE",
  "TIMING",
  "TTITLE",
  "UNDEFINE",
  "VARIABLE",
  "WHENEVER",
  "XQUERY"
};


#define NUM_COMMAND (sizeof(command)/sizeof(command[0]))
#define NUM_IGNORE (sizeof(ignore)/sizeof(ignore[0]))
#define NUM_PLSQL (sizeof(plsql)/sizeof(plsql[0]))


static gboolean is_command_word(const gchar *word)
{
	guint i;
  
	for (i=0; i < NUM_COMMAND; i++) {
		if (!g_ascii_strncasecmp(word, command[i], strlen(command[i]))) {
			return TRUE;
		}
	}
	return FALSE;
}


static gboolean is_ignore_word(const gchar *word)
{
	guint i;

	for (i=0; i < NUM_IGNORE; i++) {
		if (!g_ascii_strncasecmp(word, ignore[i], strlen(ignore[i]))) {
			return TRUE;
		}
	}
	return FALSE;
}


static gboolean is_plsql_word(const gchar *word)
{
	guint i;
  
	for (i=0; i < NUM_PLSQL; i++) {
		if (!g_ascii_strncasecmp(word, plsql[i], strlen(plsql[i]))) {
			return TRUE;
		}
	}
	return FALSE;
}


static guint skip_to_next_line(const gchar *buffer, const gint len, guint pos)
{
	while (pos < len && buffer[pos] != '\n') {
		pos++;
	}

	if (buffer[pos] == '\n') {
		pos++;
	}
	return pos;
}


static guint skip_double_quoted(const gchar *buffer, const gint len, guint pos)
{
        if (buffer[pos] != '\"') {
                return pos;
        }

        pos++;

        while (pos < len && buffer[pos] != '\"') {
                pos++;

                if (buffer[pos] == '\"' && buffer[pos-1] == '\\') {
                        pos++;
                }
        }

        if (buffer[pos] == '\"') {
                pos++;
        }
        return pos;
}

static guint skip_single_quoted(const gchar *buffer, const gint len, guint pos)
{

        if (buffer[pos] != '\'') {
                return pos;
        }

        pos++;

        while (pos < len && buffer[pos] != '\'') {
                pos++;
        }

        if (buffer[pos] == '\'') {
                pos++;

                if (buffer[pos] == '\'') {
                        pos = skip_single_quoted(buffer, len, pos);
                }
        }

        return pos;
}

static gboolean is_plsql_command(const GArray *words, gint word_num)
{
	if (!g_ascii_strncasecmp("CREATE", g_array_index(words, gchar *, word_num), 6)) {
		if (words->len > word_num + 1) {
			if (is_plsql_word(g_array_index(words, gchar *, word_num+1))) {
				return TRUE;
			}
		}
	}
	return FALSE;
}


static gboolean slash_by_itself(const char* buffer, const guint start, const guint pos)
{
	guint eol = pos+1;
	guint bol = pos-1;

	while (bol > start && buffer[bol] != '\n') {
		if (!g_ascii_isspace(buffer[bol])) {
			return FALSE;
		}
		bol--;
	}

	while (buffer[eol] != '\n' && buffer[eol] != '\0') {
		if (!g_ascii_isspace(buffer[eol])) {
			return FALSE;
		}
		eol++;
	}
	return TRUE;
}


static guint skip_space(const gchar *buffer, const gint len, guint pos)
{
	while (pos < len && g_ascii_isspace(buffer[pos])) {
		pos++;
	}
	return pos;
}


static guint skip_comments(const gchar *buffer, const gint len, guint pos)
{
	if (pos < len-2 && (buffer[pos] != '/' && buffer[pos+1] != '*')) {
		return pos;
	}

	while (pos < len-3) {
		if (buffer[pos+1] == '*' && buffer[pos+2] == '/') {
			pos += 3;
			break;
		}
		else {
			pos++;
		}
	}
	return pos;
}


static guint advance_to_cmd(const char *buffer, guint pos)
{
	while (pos < strlen(buffer) 
	      && (g_ascii_isspace(buffer[pos]) || buffer[pos]=='/')) {

		pos = skip_space(buffer, strlen(buffer), pos);
		pos = skip_comments(buffer, strlen(buffer), pos);
	}
	return pos;
}


GSList *get_statements(GtkTextView *view) 
{
	GSList *stlist=0;
	GtkTextBuffer *tb=0;
	GtkTextIter start, end;
	gchar *buffer=0;
	guint cursor=0, offset=0, ststart=0;
	stmt *stnode=0;


	tb = gtk_text_view_get_buffer(view);

	if (gtk_text_buffer_get_selection_bounds(tb, &start, &end)) {
		offset = gtk_text_iter_get_offset(&start);
	}
	else {
		gtk_text_buffer_get_bounds(tb, &start, &end);
	}

	buffer = gtk_text_buffer_get_text(tb, &start, &end, FALSE);
	if (!buffer) {
		return 0;
	}

	while (cursor < strlen(buffer)) {
		guint z=0, length=0;
		gboolean execute_by_slash=FALSE;
		GArray *word=0;
		guint wordcount=0;
		gchar delim=0;
	
		cursor = advance_to_cmd(buffer, cursor);

		word = g_strsplit_garray(&buffer[cursor], " \n\t", -1);

#if DEBUG
		fprintf(stderr, "%s:%d split at buffer[%d]\n",
			__FILE__, __LINE__, cursor);
#endif

		for (z=0; z < word->len; z++ ) {
			const gchar *lword = g_array_index(word, gchar *, z);
#if DEBUG
			fprintf(stderr, "%s:%d wordnum=%d wordcount=%d - >%s<\n",
				__FILE__, __LINE__, z, wordcount, lword);
#endif

			if (is_ignore_word(lword)) {
				cursor = skip_to_next_line(buffer, strlen(buffer), cursor);
				wordcount=-1;
				break;
			}

			if (is_command_word(g_array_index(word, gchar *, z))) {
#if DEBUG
				fprintf(stderr, "%s:%d KEYWORD: %s\n",
					__FILE__, __LINE__, lword);
#endif
				delim = ';';

				if (is_plsql_command(word, z)) {
#if DEBUG
					fprintf(stderr, "%s:%d PL_SQL DDL\n",
						__FILE__, __LINE__);
#endif
					delim = 0;
				}

				wordcount=1;
				break;
			}


			if (!g_ascii_strncasecmp(lword, "BEGIN", 5) ||
			    !g_ascii_strncasecmp(lword, "DECLARE", 7) ) {
#if DEBUG
				fprintf(stderr, "%s:%d ANON PLSQL: %s\n",
					__FILE__, __LINE__, lword);
#endif
				wordcount=1;
				break;
			}

		}

		g_strsplit_garray_free(word);
#if DEBUG
		fprintf(stderr,"%s:%d cursor=%d  delim='%c' wordcount=%d\n",
			__FILE__, __LINE__, cursor, delim, wordcount);
#endif
		if (-1 == wordcount) {
			continue;
		}

		if (!wordcount) {
			break;
		}

		for (ststart=cursor; cursor < strlen(buffer); cursor++) {

			if (buffer[cursor]=='\'') {
                                cursor = skip_single_quoted(buffer, strlen(buffer), cursor);
                        }
			
                        if (buffer[cursor]=='\"') {
                                cursor = skip_double_quoted(buffer, strlen(buffer), cursor);
                        }

			if (buffer[cursor]=='/' ) {

				switch(buffer[cursor+1]) {

				case '*':
					cursor = skip_comments(buffer, strlen(buffer), cursor);
					break;
	
				default:
					execute_by_slash = slash_by_itself(buffer, ststart, cursor);
					break;
				}
			}

			if (delim && buffer[cursor] == delim) {
				break;
			}

			if (buffer[cursor] == '/' && execute_by_slash) {
				cursor--;
				break;
			}

		}

		length = (cursor - ststart) + 1;
		stnode = (stmt *) g_malloc(sizeof(stmt));
		stnode->sql = 0;
		stnode->offset = ststart + offset;

		if (length) {
			stnode->sql = (gchar *) g_malloc(length);
			g_snprintf(stnode->sql, (gulong)length, "%s", &buffer[ststart] );
		}

		stlist = g_slist_append(stlist, (gpointer) stnode);
    
#if DEBUG
		fprintf(stderr,"%s:%d offset: %d  ststart: %d  cursor: %d\n>>>%s<<<\n",
			__FILE__, __LINE__, offset, ststart, cursor, stnode->sql);
#endif
		cursor++;
		if (execute_by_slash) {
			cursor++;
		}
	}
  
	g_free(buffer);
	return(stlist);
}






