/* This program is free software; you can redistribute it and/or */
/* modify it under the terms of the GNU General Public License */
/* as published by the Free Software Foundation; either version 2 */
/* of the License, or (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <oci.h>
#include <glib.h>

#include "orautil.h"
#include "command_tab.h"
#include "gutil.h"
#include "trace.h"
#include "interface.h"
#include "support.h"
#include "appconfig.h"

extern GtkWidget       *window1;
extern OCIEnv          *envhp;

static GSList          *stat_list=0;
static unsigned int    *trace_vals=0, *mark_vals=0; 


static gint            trace_sid = 0;
static OCIStmt         *trace_sth = (OCIStmt *)0;
static GString         *trsql = NULL;

static GtkWidget       *traceconfWindow=0;

static void construct_trace_query(const gchar *stat_csv)
{
	gchar **stat_arr=0;
	guint x;
	GString *val_str=g_string_new("");
	gchar *default_stats[] = {
		"execute count",
		"recursive calls",
		"db block gets",
		"db block changes",
		"consistent gets",
		"consistent changes",
		"session logical reads",
		"physical reads",
		"physical writes",
		"redo size",
		"enqueue requests",
		"enqueue waits",
		"bytes sent via SQL*Net to client",
		"bytes received via SQL*Net from client",
		"SQL*Net roundtrips to/from client",
		"sorts (memory)",
		"sorts (disk)",
		"sorts (rows)",
		"CPU used by this session"
	};

#define NUM_DEFSTAT (sizeof(default_stats) / sizeof(default_stats[0]))


	if (stat_list) {
		g_slist_free(stat_list);
		stat_list=0;
	}

	if (stat_csv) { 

		stat_arr = g_strsplit_set(stat_csv, ",", -1);

		for(x=0; stat_arr[x]; x++) 
			stat_list = g_slist_append(stat_list, g_strdup(stat_arr[x]) );

		g_strfreev(stat_arr);

	}
	else {

		for (x=0; x < NUM_DEFSTAT; x++) {
			gchar sappend[256];
    
			stat_list = g_slist_append(stat_list, g_strdup(default_stats[x]) );
     
			g_snprintf(sappend, sizeof(sappend), "%s%s", x > 0 ? "," : "", default_stats[x]);
			val_str = g_string_append(val_str, sappend);
		}
    
		g_object_set_data(G_OBJECT(window1), "trace_list", val_str->str);
		g_string_free(val_str, FALSE);
	}

	trace_vals = g_new0(unsigned int, g_slist_length(stat_list));
	mark_vals  = g_new0(unsigned int, g_slist_length(stat_list));


	if(trsql)
		g_string_free(trsql, TRUE);

	trsql = g_string_new("SELECT\n");
	for(x=0; x < g_slist_length(stat_list); x++) {
		gchar     stat_decode[128];
    
		snprintf(stat_decode, sizeof(stat_decode), "  sum(decode(s.name, '%s', m.value, 0)),\n", (gchar *) g_slist_nth_data(stat_list, x) );
		trsql = g_string_append(trsql, stat_decode);
	}

	trsql = g_string_erase(trsql, trsql->len-2, 2);
	trsql = g_string_append(trsql, "\nfrom v$sesstat m, v$statname s where s.statistic# = m.statistic# and m.sid = :p1");

}



void session_trace_init(OCIEnv *envhp,
			OCIError *errhp,
			gint session_id,
			const gchar *trace_list)
{
	trace_sid = session_id;

	construct_trace_query(trace_list);

	if (trace_sth) {
		OCIHandleFree(trace_sth, OCI_HTYPE_STMT);
	}

	OCIHandleAlloc(envhp, (dvoid **)&trace_sth, (ub4)OCI_HTYPE_STMT, 0, (dvoid **)0);
}


static void  session_trace_fetch(OCISvcCtx *svchp,
				 OCIError  *errhp)
{
	guint x;
	OCIDefine *defnp = (OCIDefine *)0;
	OCIBind *bhp = (OCIBind *)0;

	if ((ociErr(errhp, ociPrepare(trace_sth, errhp, trsql->str)) == OCI_SUCCESS)) {

		ociErr(errhp, OCIBindByPos((OCIStmt *) trace_sth, 
					   (OCIBind **) &bhp, 
					   (OCIError *) errhp, 
					   (ub4) 1, 
					   (dvoid *)&trace_sid, (sb4) sizeof(trace_sid), 
					   (ub2)SQLT_INT, 
					   (dvoid *)0, (ub2 *)0, (ub2 *)0, (ub4)0, (ub4 *)0, 
					   (ub4)OCI_DEFAULT)
			);

		for (x=0; x < g_slist_length(stat_list); x++) 
			OCIDefineByPos(trace_sth,
				       &defnp,
				       errhp,
				       (ub4) x+1, 
				       (dvoid *) &trace_vals[x], (sb4) sizeof(unsigned int),
				       (ub2) SQLT_INT,
				       (dvoid *) 0,
				       (ub2 *) 0,
				       (ub2 *) 0,
				       (ub4) OCI_DEFAULT);

		ociErr(errhp, ociExecute(svchp, trace_sth, errhp, 1));

#if DEBUG
		for(x=0; x < g_slist_length(stat_list); x++) 
			fprintf(stderr, "trace.c: %3d: %-48s %d\n", x, (gchar *) g_slist_nth_data(stat_list, x), trace_vals[x]);
#endif

	}
}




void  session_trace_mark(OCISvcCtx *svchp, 
			 OCIError *errhp)
{
	guint x;
	OCIDefine *defnp = (OCIDefine *)0;
	OCIBind *bhp   = (OCIBind *)0;

	if ((ociErr(errhp, ociPrepare(trace_sth, errhp, trsql->str)) == OCI_SUCCESS)) {

		ociErr(errhp, OCIBindByPos((OCIStmt *) trace_sth, 
					   (OCIBind **) &bhp, 
					   (OCIError *) errhp, 
					   (ub4) 1, 
					   (dvoid *)&trace_sid, (sb4) sizeof(trace_sid), 
					   (ub2)SQLT_INT, 
					   (dvoid *)0, (ub2 *)0, (ub2 *)0, (ub4)0, (ub4 *)0, 
					   (ub4)OCI_DEFAULT)
			);

		for(x=0; x < g_slist_length(stat_list); x++) 
			OCIDefineByPos(trace_sth,
				       &defnp,
				       errhp,
				       (ub4) x+1, 
				       (dvoid *) &mark_vals[x], (sb4)sizeof(unsigned int),
				       (ub2) SQLT_INT,
				       (dvoid *) 0,
				       (ub2 *) 0,
				       (ub2 *) 0,
				       (ub4) OCI_DEFAULT);

		ociErr(errhp, ociExecute(svchp, trace_sth, errhp, 1));
	}
}


void  session_trace_display(OCISvcCtx *svchp,
			    OCIError  *errhp,
			    gchar     *viewname,
			    gdouble   elapsed,
			    gint      rows)
{
	GtkTreeView        *list = GTK_TREE_VIEW(lw(viewname));
	GtkCellRenderer    *crend;
	GtkTreeViewColumn  *column;
	GtkTreeIter        iter;
	GtkTreeStore       *store;
	GList              *lcols;
  
	guint              x;
	const gchar        *font = config_get("listfont");

	GIOChannel         *fio=0;
	const gchar        *query;
	gchar              file_buf[256];
	gboolean           to_file, with_query, with_plan, append;
	gsize              written;


	to_file = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lw("traceFileCheck")));
	with_query = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lw("traceWithQueryCheck")));
	with_plan = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lw("traceWithPlanCheck")));
	append = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lw("traceAppendCheck")));
	query = g_object_get_data(G_OBJECT(lw("resList")), "last_query");

	if (to_file) {
		gchar   *outfile = (gchar *) g_object_get_data(G_OBJECT(lw("traceFileChooserButton")), "filename");

		if (outfile) {
			GError *gerror=0;
      
			fio = g_io_channel_new_file(outfile, gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lw("traceAppendCheck"))) ? "a" : "w", &gerror);

			if(gerror) {
				gchar   msg[256];

				to_file=FALSE;
				g_snprintf(msg, sizeof(msg), "Error opening trace file for write: %s\n", gerror->message);
				gtk_label_set_text(GTK_LABEL(lw("statusLabel")), msg);
				g_error_free(gerror);
			} else {
	
				if(with_query) {
					if(query) {
						g_io_channel_write_chars(fio, "/**** QUERY ****/\n", -1, &written, (GError **) NULL );
						g_io_channel_write_chars(fio, query, -1, &written, (GError **) NULL );
						g_io_channel_write_chars(fio, "\n", -1, &written, (GError **) NULL );
					}
				}

	
				if(with_plan) {
					OCIStmt    *sth=0;
					OCIDefine  *defnp=0;
					gchar      *expsql, plan_buf[512];

					//g_io_channel_write_chars(fio, "/****  PLAN  ****/", -1, &written, (GError **) NULL );

					OCIHandleAlloc(envhp, (dvoid **)&sth, (ub4)OCI_HTYPE_STMT, 0, (dvoid **)0);

					expsql = g_strconcat("explain plan for ", query, NULL);

					if( (ociErr(errhp, ociPrepare(sth, errhp, expsql)) == OCI_SUCCESS)) {
						ociErr(errhp, ociExecute(svchp, sth, errhp, 1));
					}
					g_free(expsql);

					if( (ociErr(errhp, ociPrepare(sth, errhp, "select * from table(dbms_xplan.display)")) == OCI_SUCCESS)) {
      
						OCIDefineByPos(sth,
							       &defnp,
							       errhp,
							       (ub4) 1, 
							       (dvoid *) plan_buf, (sb4)sizeof(plan_buf)+1,
							       (ub2) SQLT_STR,
							       (dvoid *) 0,
							       (ub2 *) 0,
							       (ub2 *) 0,
							       (ub4) OCI_DEFAULT);

						ociErr(errhp, ociExecute(svchp, sth, errhp, 1));
						while( OCIStmtFetch2(sth, errhp, 1, OCI_FETCH_NEXT, 0, OCI_DEFAULT) == OCI_SUCCESS) {
							g_io_channel_write_chars(fio, plan_buf, -1, &written, (GError **) NULL );
							g_io_channel_write_chars(fio, "\n", -1, &written, (GError **) NULL );
						}
						g_io_channel_write_chars(fio, "\n", -1, &written, (GError **) NULL );
					}
					OCIHandleFree(sth, OCI_HTYPE_STMT);
				}
			}
		}
	}

	store = GTK_TREE_STORE(gtk_tree_view_get_model(list));
	if(store)
		gtk_tree_store_clear(GTK_TREE_STORE(store)); 

	lcols = gtk_tree_view_get_columns(list);
	g_list_foreach(lcols, (GFunc) delcolumn, (gpointer) list);
	g_list_free(lcols);

	crend = gtk_cell_renderer_text_new();

	if(font)
		g_object_set(G_OBJECT(crend), "font", font, NULL);

	g_object_set(G_OBJECT(crend), "ypad", 0, NULL);

	column = gtk_tree_view_column_new_with_attributes("Value", crend, "text", 0, NULL);
	gtk_tree_view_append_column(list, column);
	column = gtk_tree_view_column_new_with_attributes("Statistic", crend, "text", 1, NULL);
	gtk_tree_view_append_column(list, column);

	store = gtk_tree_store_new(2, G_TYPE_INT, G_TYPE_STRING);
  
	if(fio) {
		g_io_channel_write_chars(fio, "Value             Statistic\n---------------- ----------------------------\n", -1, &written, (GError **) NULL );
	}

	session_trace_fetch(svchp, errhp);

	if( ociErr(errhp, ociExecute(svchp, trace_sth, errhp, 1)) == OCI_SUCCESS) {

		for(x=0; x < g_slist_length(stat_list); x++) {

			gtk_tree_store_append(store, &iter, NULL);
			gtk_tree_store_set(store, &iter, 0, trace_vals[x] - mark_vals[x], 1, g_slist_nth_data(stat_list, x), -1);


			if(fio) {
	
				snprintf(file_buf, sizeof(file_buf), "%16d  %s\n", trace_vals[x] - mark_vals[x], (gchar *)g_slist_nth_data(stat_list, x) );
				g_io_channel_write_chars(fio, file_buf, -1, &written, (GError **) NULL );
			}

		}

		if(elapsed != -1) {
			g_snprintf(file_buf, sizeof(file_buf), "%.03f Elapsed", elapsed);
			gtk_tree_store_append(store, &iter, NULL);
			gtk_tree_store_set(store, &iter, 1, file_buf, -1);

			if(fio) {
	
				snprintf(file_buf, sizeof(file_buf), "%.03f Elapsed\n%d Rows\n\n\n\n", elapsed, rows );
				g_io_channel_write_chars(fio, file_buf, -1, &written, (GError **) NULL );
			}

		}
	}

	gtk_tree_view_set_model(list, GTK_TREE_MODEL(store) );

	if(fio) {
		g_io_channel_shutdown(fio, TRUE, (GError **)NULL);
		g_io_channel_unref(fio);
	}

}



void trace_config_cancel()
{
	if (traceconfWindow) {
		gtk_widget_destroy(traceconfWindow);
		traceconfWindow = 0;
	}
}


void trace_config_ok()
{
	GtkTreeModel  *model = gtk_tree_view_get_model(GTK_TREE_VIEW(lookup_widget(traceconfWindow, "traceconfView")));
	GtkTreeIter   iter;
	guint         ret;
	GString       *val_str=g_string_new("");
	guint         n=0;


	for (ret=gtk_tree_model_get_iter_first(model, &iter); ret!=FALSE; ret=gtk_tree_model_iter_next(model, &iter) ) {
		GtkTreeIter   child;
		guint         cret=TRUE;
		gchar         *statname;
		gboolean      togg;

		gtk_tree_model_iter_children(model, &child, &iter);

		for(; cret!=FALSE; cret=gtk_tree_model_iter_next(model, &child) ) {

			gtk_tree_model_get(model, &child, 0, &statname, 2, &togg, -1);
      
			if(togg) {
				gchar    sappend[256];
	
				g_snprintf(sappend, sizeof(sappend), "%s%s", n++ > 0 ? "," : "", statname);
				val_str = g_string_append(val_str, sappend);
			}

			g_free(statname);
      
		}
	}
 
	construct_trace_query(val_str->str);

	config_set("trace_list", val_str->str);
	g_string_free(val_str, FALSE);

	if(traceconfWindow) {
		gtk_widget_destroy(traceconfWindow);
		traceconfWindow = NULL;
	}
}



static void item_toggled (GtkCellRendererToggle *cell,
			  gchar *path_str,
			  gpointer data)
{
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(lookup_widget(traceconfWindow, "traceconfView")));
	GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
	GtkTreeIter iter;
	gboolean toggle_item;
	gchar *statname=0;


	gtk_tree_model_get_iter(model, &iter, path);
	gtk_tree_model_get(model, &iter, 0, &statname, 2, &toggle_item, -1);

	toggle_item ^= 1;

	gtk_tree_store_set (GTK_TREE_STORE (model), &iter, 2, toggle_item, -1);

	g_free(statname);
	gtk_tree_path_free (path);
}




void  trace_config_pop(OCIEnv *envhp, OCISvcCtx *svchp, OCIError *errhp) 
{
	GtkTreeView *list;
	GtkCellRenderer *crend, *trend;
	GtkTreeViewColumn *column;
	GtkTreeIter iter, piter;
	GtkTreeStore *store;
	OCIStmt *sth = (OCIStmt *)0;
	OCIDefine *defnp = (OCIDefine *)0;
	gchar class[32], statname[128];
	guint branch;
	const gchar  *font = config_get("listfont");
	gchar *sql = "\
  select \
  decode(class, \
   1, 'User', \
   2, 'Redo', \
   4, 'Enqueue', \
   8, 'Cache', \
   16, 'OS', \
   32, 'Parallel', \
   40, 'RAC', \
   64, 'SQL', \
   72, 'Buffer', \
   128, 'Debug', \
   'unknown'),\
  name, \
  decode(nvl(lag(class) over (order by class), 0), class, 0, 1)\
  from v$statname order by class, name";


	if(traceconfWindow)
		return;

	traceconfWindow = create_traceconfWindow();
	list = GTK_TREE_VIEW(lookup_widget(traceconfWindow, "traceconfView"));

	crend = gtk_cell_renderer_text_new();
	trend = gtk_cell_renderer_toggle_new();
  
	if(font)
		g_object_set(G_OBJECT(crend), "font", font, NULL);

	g_object_set(G_OBJECT(crend), "ypad", 0, NULL);

	column = gtk_tree_view_column_new_with_attributes("Class", crend, "text", 0, NULL);
	gtk_tree_view_append_column(list, column);
	column = gtk_tree_view_column_new_with_attributes("Selected", trend, "activatable", 1, "active", 2, NULL);
	gtk_tree_view_append_column(list, column);
	g_signal_connect (trend, "toggled", G_CALLBACK (item_toggled), (gpointer) NULL);

	store = gtk_tree_store_new(3, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
  

	OCIHandleAlloc(envhp, (dvoid **)&sth, (ub4)OCI_HTYPE_STMT, 0, (dvoid **)0);
	if( (ociErr(errhp, ociPrepare(sth, errhp, sql)) == OCI_SUCCESS)) {

		OCIDefineByPos(sth,
			       &defnp,
			       errhp,
			       (ub4) 1, 
			       (dvoid *) &class, (sb4)sizeof(class)+1,
			       (ub2) SQLT_STR,
			       (dvoid *) 0,
			       (ub2 *) 0,
			       (ub2 *) 0,
			       (ub4) OCI_DEFAULT);
		OCIDefineByPos(sth,
			       &defnp,
			       errhp,
			       (ub4) 2, 
			       (dvoid *) &statname, (sb4)sizeof(statname)+1,
			       (ub2) SQLT_STR,
			       (dvoid *) 0,
			       (ub2 *) 0,
			       (ub2 *) 0,
			       (ub4) OCI_DEFAULT);
		OCIDefineByPos(sth,
			       &defnp,
			       errhp,
			       (ub4) 3, 
			       (dvoid *) &branch, (sb4)sizeof(branch),
			       (ub2) SQLT_INT,
			       (dvoid *) 0,
			       (ub2 *) 0,
			       (ub2 *) 0,
			       (ub4) OCI_DEFAULT);
    
		ociErr(errhp, ociExecute(svchp, sth, errhp, 0));
		while( OCIStmtFetch2(sth, errhp, 1, OCI_FETCH_NEXT, 0, OCI_DEFAULT) == OCI_SUCCESS) {
			gboolean    checked=FALSE;
			guint       x;

			if( branch ) {
				gtk_tree_store_append(store, &piter, NULL);
				gtk_tree_store_set(store, &piter, 0, class, -1);
			}
   
			for(x=0; x < g_slist_length(stat_list); x++) {
				if(!g_ascii_strncasecmp(statname, g_slist_nth_data(stat_list, x), strlen(statname))) {
					checked = TRUE;
					break;
				}
			}

			gtk_tree_store_append(store, &iter, &piter);
			gtk_tree_store_set(store, &iter, 0, statname, 1, TRUE, 2, checked, -1);
		}
	}

	gtk_tree_view_set_model(list, GTK_TREE_MODEL(store) );
	gtk_widget_show(traceconfWindow);
}





void create_traceFileChooser (void)
{
	GtkWidget *saveChooser;
	GtkWidget *vbox;
	GtkWidget *action_area;
	GtkWidget *saveFileCancel;
	GtkWidget *saveFileButton;
	gchar     *filename=0, *basename=0;
	GtkTooltips *tip = gtk_tooltips_new();

	saveChooser = gtk_file_chooser_dialog_new ("Select trace output file...", NULL, GTK_FILE_CHOOSER_ACTION_SAVE, NULL, NULL);
	gtk_window_set_type_hint (GTK_WINDOW (saveChooser), GDK_WINDOW_TYPE_HINT_DIALOG);

	vbox = GTK_DIALOG (saveChooser)->vbox;
	gtk_widget_show (vbox);

	action_area = GTK_DIALOG (saveChooser)->action_area;
	gtk_widget_show (action_area);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (action_area), GTK_BUTTONBOX_END);

	saveFileCancel = gtk_button_new_from_stock ("gtk-cancel");
	gtk_widget_show (saveFileCancel);
	gtk_dialog_add_action_widget (GTK_DIALOG (saveChooser), saveFileCancel, GTK_RESPONSE_CANCEL);
	GTK_WIDGET_SET_FLAGS (saveFileCancel, GTK_CAN_DEFAULT);

	saveFileButton = gtk_button_new_from_stock ("gtk-apply");
	gtk_widget_show (saveFileButton);
	gtk_dialog_add_action_widget (GTK_DIALOG (saveChooser), saveFileButton, GTK_RESPONSE_OK);
	GTK_WIDGET_SET_FLAGS (saveFileButton, GTK_CAN_DEFAULT);

	filename = g_object_get_data(G_OBJECT(lw("traceFileChooserButton")), "filename");

	if(filename) {
		gchar *dirname = g_path_get_dirname(filename);

		if(dirname) {
			gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(saveChooser), dirname);
			g_free(dirname);
		}
		gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(saveChooser), filename);
	}
	else {
		const gchar   *pagelabel = gtk_label_get_text(GTK_LABEL(command_tab_get_page_label()));

		if(pagelabel) {
			gchar   filelabel[256];

			g_snprintf(filelabel, sizeof(filelabel), "%s.trc", pagelabel);
			gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(saveChooser), filelabel);
		}
		else {
			gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(saveChooser), "gtko_trace.trc");
		}
	}

	gtk_widget_grab_default (saveFileButton);

	switch(gtk_dialog_run (GTK_DIALOG (saveChooser)) ) {

	case GTK_RESPONSE_OK:

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(saveChooser) );

		if(filename) {
  
			basename = g_path_get_basename(filename);
			if(basename)
				gtk_label_set_text(GTK_LABEL(lw("traceFileChooserLabel")), basename);

			g_object_set_data_full(G_OBJECT(lw("traceFileChooserButton")), "filename", (gpointer) filename, g_free);
			g_free(basename);

			gtk_tooltips_set_tip(tip, lw("traceFileChooserButton"), filename, NULL);
			gtk_widget_destroy(saveChooser);
		}

		break;

	default:
		gtk_widget_destroy(saveChooser);

		break;
	}

	return;
}
